"""This module provides several functions to render Collatz sequences into various displaying formats"""

import io
import pathlib

import matplotlib as mpl
import matplotlib.pyplot as plt
import networkx as nx
import numpy as np

from syracuse import Syracuse, CompressedSyracuse

def single_sequence_to_dot_string(syr_sequence:Syracuse, converter:str = "native") -> str:
	"""Create a string containing a [Graphviz](https://graphviz.org/) dot format representing a single Collatz sequence, that can be read by Graphviz to render into various image formats.
	
	It is possible to choose among several "converters" to create dot strings from the Collatz sequence. The "native" one (the default) does not need any external module. It is inspirated by [https://en.wikipedia.org/wiki/File:Collatz-graph-300.svg](https://en.wikipedia.org/wiki/File:Collatz-graph-300.svg). The "pydot" one uses the `pydot` library, and creates the dot string from the internal Networkx graph.
	
	Examples:
		>>> import syracuse
		>>> import syracuse.drawing
		>>>
		>>> # Native converter
		>>> syr = syracuse.Syracuse(6)
		>>> collatz6 = syracuse.drawing.single_sequence_to_dot_string(syr)
		>>> print(collatz6)
		digraph {
		6 -> 3 -> 10 -> 5 -> 16 -> 8 -> 4 -> 2 -> 1;
		}
		>>>
		>>> # pydot converter
		>>> syr = syracuse.Syracuse(10)
		>>> collatz10 = syracuse.drawing.single_sequence_to_dot_string(syr, "pydot")
		>>> print(collatz10)
		strict digraph {
		10;
		5;
		16;
		8;
		4;
		2;
		1;
		10 -> 5;
		5 -> 16;
		16 -> 8;
		8 -> 4;
		4 -> 2;
		2 -> 1;
		}
	
	Parameters:
		syr_sequence:
			Sequence to be rendered
		converter:
			The converter used to create the dot string. The available values are: "native" or "pydot"
	
	Returns:
		A Graphviz dot format representation of the sequence graph
	"""
	if not isinstance(syr_sequence, Syracuse):
		raise ValueError("`syr_sequence` must be a Syracuse object")
	else:
		if converter.lower() == "pydot":
			output_str = nx.nx_pydot.to_pydot(syr_sequence.graph).to_string()
		else:
			output = io.StringIO()
			output.write("digraph {\n")
			for node in syr_sequence:
				if node == 1:
					break
				else:
					output.write(str(node) + " -> ")
			output.write("1;\n}\n")
			output_str = output.getvalue()
			output.close()
		return output_str

def range_sequences_to_dot_string(compressed:bool = False, limit:int = 10, orientation:str = "portrait", excludes:list[int] = [], colored:bool = False, converter:str = "native") -> str:
	"""Create a string containing a [Graphviz](https://graphviz.org/) dot format representing a range of Collatz sequences, that can be read by Graphviz to render into various image formats.
	
	It is possible to choose among several "converters" to create dot strings from the Collatz sequence. The "native" one (the default) does not need any external module. It is inspirated by [https://en.wikipedia.org/wiki/File:Collatz-graph-300.svg](https://en.wikipedia.org/wiki/File:Collatz-graph-300.svg). The "pydot" one uses the `pydot` library, and creates the dot string from the Networkx global graph.
	
	Examples:
		>>> import syracuse
		>>> import syracuse.drawing
		>>> collatz_range15 = syracuse.drawing.range_sequences_to_dot_string(limit = 15)
		>>> print(collatz_range15)
		digraph {
		1;
		2 -> 1;
		3 -> 10 -> 5 -> 16 -> 8 -> 4 -> 2;
		4;
		5;
		6 -> 3;
		7 -> 22 -> 11 -> 34 -> 17 -> 52 -> 26 -> 13 -> 40 -> 20 -> 10;
		8;
		9 -> 28 -> 14 -> 7;
		10;
		11;
		12 -> 6;
		13;
		14;
		15 -> 46 -> 23 -> 70 -> 35 -> 106 -> 53 -> 160 -> 80 -> 40;
		}
		>>> collatz_range10_horizontal_colored = syracuse.drawing.range_sequences_to_dot_string(orientation = "landscape", colored = True)
		>>> print(collatz_range10_horizontal_colored)
		digraph {
		node [colorscheme=spectral10]
		1 [color=1]
		2 [color=1]
		3 [color=1]
		10 [color=3]
		5 [color=2]
		16 [color=4]
		8 [color=2]
		4 [color=1]
		6 [color=2]
		7 [color=2]
		22 [color=5]
		11 [color=3]
		34 [color=7]
		17 [color=4]
		52 [color=10]
		26 [color=6]
		13 [color=3]
		40 [color=9]
		20 [color=5]
		9 [color=2]
		28 [color=6]
		14 [color=3]
		rankdir="LR"
		1;
		2 -> 1;
		3 -> 10 -> 5 -> 16 -> 8 -> 4 -> 2;
		4;
		5;
		6 -> 3;
		7 -> 22 -> 11 -> 34 -> 17 -> 52 -> 26 -> 13 -> 40 -> 20 -> 10;
		8;
		9 -> 28 -> 14 -> 7;
		10;
		}
	
	Parameters:
		compressed:
			indicates weither one deals with either compressed or "normal" sequences
		limit:
			the upper bound of the range values from whom the sequences are generated
		orientation:
			orientation of the rendered graph. Possible values: `landscape` or `portrait`
		excludes:
			list of sequences excluded in the range. No effect for members greater than `limit`.
		colored:
			If `True`, the rendered graph will have gradient colors, depending on the value of each node
		converter:
			The converter used to create the dot string. The available values are: "native" or "pydot". Please notice that this parameter is only relevant if colored is `False` (if colored is `True`, the native converter is used).
	
	Returns:
		A Graphviz dot format representation of the sequences range graph
	"""
	if not isinstance(limit, int):
		raise TypeError("`limit` must be an integer")
	else:
		if converter.lower() == "pydot" and not colored:
			output_str = nx.nx_pydot.to_pydot(Syracuse.generate_global_graph(max_initial_value=limit, min_initial_value=1, excludes=excludes)).to_string()
		else:
			output = io.StringIO()
			output.write("digraph {\n")
			max_all = 1
			seq_range = [seq_start for seq_start in range(1, limit+1) if seq_start not in excludes]
			if colored:
				explored = set([1])
				output.write("node [colorscheme=spectral10]\n")
				min_color_range = 1
				max_color_range = 10
				max_list = []
				for n in seq_range:
					max_list.append(CompressedSyracuse(n).max if compressed else Syracuse(n).max)
				max_all = max(max_list)
				output.write(f"1 [color={min_color_range}]\n")
				for n in seq_range:
					syr_seq = CompressedSyracuse(n) if compressed else Syracuse(n)
					for node in syr_seq:
						if node in explored:
							break
						else:
							if node == max_all:
								output.write(str(max_all) + f"[color={max_color_range}]\n")
							else:
								output.write(str(node) + f" [color={node//(max_all//max_color_range)+1}]" + "\n")
							explored.add(node)
			if orientation.lower() == "landscape":
				output.write('rankdir="LR"\n')
			explored = set([1])
			for n in seq_range:
				syr_seq = CompressedSyracuse(n) if compressed else Syracuse(n)
				for node in syr_seq:
					if node in explored:
						output.write(str(node) + ";\n")
						break
					else:
						output.write(str(node) + " -> ")
						explored.add(node)
			output.write("}\n")
			output_str = output.getvalue()
			output.close()
		return output_str

def single_sequence_to_matplotlib_figure(syr_sequence:Syracuse, test:bool = False) -> mpl.figure.Figure:
	"""Create a [Matplotlib](https://matplotlib.org/) Figure object representing the evolution of the successive values of a particular Collatz sequence.
	
	Parameters:
		syr_sequence:
			Sequence to be rendered
		test:
			if `True`, displays the graph immediately (for testing purposes)
	
	Returns:
		matplotlib.figure.Figure: A representation of the successive values of the Collatz sequence
	"""
	if not isinstance(syr_sequence, Syracuse):
		raise ValueError("`syr_sequence` must be a Syracuse object")
	else:
		fig, ax = plt.subplots()
		ax.plot(syr_sequence.total_stopping_sequence)
		ax.set_title(f"Syracuse sequence of {syr_sequence.initial_value}")
		ax.annotate(f"Total stopping time: {syr_sequence.total_stopping_time}",
					xy=(syr_sequence.total_stopping_time,1), xycoords='data',
					xytext=(-15,25), textcoords="offset points",
					arrowprops=dict(facecolor='black', shrink=0.05),
					horizontalalignment='right', verticalalignment='bottom')
		ax.annotate(f"Max value: {syr_sequence.max}",
					xy=(syr_sequence.max_rank,syr_sequence.max), xycoords='data',
					xytext=(25,-15), textcoords="offset points",
					arrowprops=dict(facecolor='black', shrink=0.05),
					horizontalalignment='left', verticalalignment='top')
		
		# Only for debug !
		if test:
			plt.show() # Display the graph
		
		return fig

def range_sequences_distribution_to_matplotlib_figure(max_initial_value:int, min_initial_value:int = 1, statistic:bool = True, test:bool = False) -> mpl.figure.Figure:
	"""Create a [Matplotlib](https://matplotlib.org/) Figure object representing the distribution of the successive values of a range of Collatz sequences.
	
	Parameters:
		min_initial_value:
			The minimal initial value of the proceeded sequences
		max_initial_value:
			The maximal initial value of the proceeded sequences
		statistic:
			If `True`, displays some statistical charateristics on the returned figure
		test:
			if `True`, displays the graph immediately (for testing purposes)
	
	Returns:
		matplotlib.figure.Figure: A bar graph representation of the distribution
	
	Raises:
		ValueError:
			Raises if `min_initial_value` or `max_initial_value` are not strictly positive integers, or if `max_initial_value` < `min_initial_value`.
	"""
	if not isinstance(max_initial_value, int) or (max_initial_value <= 0):
			raise ValueError("max_initial_value must be a strictly positive integer")
	elif not isinstance(min_initial_value, int) or (min_initial_value <= 0):
		raise ValueError("min_initial_value must be a strictly positive integer")
	elif max_initial_value < min_initial_value:
		raise ValueError("max_initial_value must be greater than min_initial_value")
	else:
		inits = range(min_initial_value, max_initial_value+1)
		weight = dict()
		members = set()
		repeatable_members = list()
		for initial_value in inits:
			for value in Syracuse(initial_value, populate_global_graph = True).total_stopping_sequence:
				repeatable_members.append(value)
				members.add(value)
				if value in weight:
					weight[value] += 1
				else:
					weight[value] = 1
		x = sorted(list(members))
		y = [weight[value] for value in x]
		
		if statistic:
			# Statistical characteristics
			median = np.median(repeatable_members) # median
			average = np.average(x, weights = y) # weighted average
			mean = np.mean(repeatable_members) # arithmetic mean
			variance = np.var(repeatable_members) # variance
			deviation = np.std(repeatable_members) # standard deviation
			
			textstr = "\n".join((
				f"median={median:.2f}",
				f"weight average={average:.2f}",
				r"$\mu=%.2f$" % (mean,),
				f"variance={variance:.2f}",
				r"$\sigma=%.2f$" % (deviation,)))
		
		fig, ax = plt.subplots()
		ax.bar(x, y, width=1, linewidth=0)
		ax.set_title(f"Distribution of the Syracuse({min_initial_value})-Syracuse({max_initial_value}) sequences")
		if statistic:
			props = dict(boxstyle="round", facecolor="wheat", alpha=0.5)
			ax.text(0.05, 0.95, textstr, transform=ax.transAxes, fontsize=14, verticalalignment="top", bbox=props)
		
		# Only for debug !
		if test:
			plt.show() # Display the graph
		
		return fig
		
	


"""This module provides two classes to represent a full or a compressed form of Collatz sequence."""

import itertools
from multiprocessing import Pool, Manager, Process, Queue
import os
from typing import Tuple

import more_itertools
import networkx as nx

class Syracuse():
	"""
	A (full) Collatz sequence instanciated with a given initial value.
	
	The instance is iterable, and gives the successive values of the Collatz sequence.
	
	!!! example
	
		```python linenums="1"
		syracuse = Syracuse(1000)
		for value in syracuse:
			print(value)
			if value == 1:
				break
		```
	
	Be aware that a Collatz sequence is infinite: once it reaches 1 (and that is apparently the case for all tested initial values until now !), the cycle 4,2,1 returns indefinitely. So make sure you have implemented a stopping condition when you iterate over the sequence.
	
	For each sequence an underlying, directional graph is implemented, and is populated the first time several attributes are read (like `total_stopping_time`, `max` and of course `graph` and `graph_view`), resulting an additionnal computing duration, that can be noticeable on "big" sequences. But the duration becomes imperceptible next times those attributes are read.
	
	Furthermore, a "global" graph can be optionaly initialized, and is shared with all instances. It is basically useful when working on a large number of sequences, or on same sequences repeatedly. If a global graph population is asked for a sequence instantiation, then the local graph is deduced from it only when needed, and stored in a class attribute for further use. Be aware that the global graph activation generates a significant amount of additional memory storage, but on the other side the computation speed is generally highly improved.
	
	Parameters:
		initial:
			Initial value of the Collatz sequence. Must be strictly positive.
		populate_global_graph:
			If `True`, a global graph is built, and populated with the graph of the current sequence. Default is `False`
	
	Attributes:
		initial_value:
			Reminder of the initial value given for the sequence initialisation. **read-only**
		stopping_time:
			The smallest rank such that the value becomes lower than the initial one. **read-only**
		total_stopping_time:
			The smallest rank such that the value equals 1. **read-only**
		total_stopping_sequence:
			Tuple of the sequence from the initial value, to 1. **read-only**
		max:
			Maximum value of the sequence. **read-only**
		max_rank:
			Rank of the maximum value of the sequence. **read-only**
		graph_view:
			A dict-like structure representing the graph of the successive members of the current sequence. **read-only**
		graph (networkx.Digraph):
			The underlying, directional graph for the current Collatz sequence. **read-only**
		global_graph (networkx.Digraph):
			The global graph shared with all sequences, gathering in the same graph all sequences' graphs instanciated with `populate_global_graph = True`. **read-only**
		single_graphs (dict[networkx.Digraph]):
			Store the graphs generated for a single sequence to avoid repeated computations of the same graph (used only if `populate_global_graph` is `True`)
		
	Raises:
		ValueError:
			Raises if `initial` is not a strictly positive integer.
	"""
	
	global_graph:nx.DiGraph = nx.DiGraph()
	single_graphs:dict[nx.DiGraph] = {}
	
	
	def __init__(self, initial:int, populate_global_graph:bool = False):
		# syracuse instance initialisation
		if not isinstance(initial, int) or (initial <= 0):
			raise ValueError
		else:
			self._initial = initial
			self._stopping_time = 0
			self._graph = nx.DiGraph()
			self._local_graph_completed = False # Flag switching to True when the instance's graph is complete
			self._populate_global_graph = populate_global_graph
			if self._populate_global_graph:
				self._generate_graph()
	
	def __iter__(self):
		# The Syracuse instance is iterable, and gives the successive values of the Collatz sequence
		Un = self._initial
		yield Un # Don't forget to yield the initial value !
		while True:
			Un = Un*3+1 if Un%2 else Un//2
			yield Un
	
	@property
	def initial_value(self) -> int:
		# Reminder of the initial value given for the sequence initialisation (read-only attribute)
		return self._initial
	
	@property
	def stopping_time(self) -> int:
		# The smallest rank such that the value becomes lower than the initial one (read-only attribute)
		#
		# The Collatz conjecture asserts that this rank is finite for every initial value strictly greater than 1. Postulating this, if `initial_value` == 1, then `stopping_time` returns 0
		
		if self._stopping_time == 0:
			if self._initial == 1:
				return self._stopping_time
			else:
				for index, value in enumerate(self):
					if value < self._initial:
						self._stopping_time = index
						return self._stopping_time
						break
		else:
			return self._stopping_time

	@property
	def total_stopping_time(self) -> int:
		# The smallest rank such that the value equals 1 (read-only attribute)
		#
		# The Collatz conjecture asserts that this rank is finite for every initial value.
		
		return len(self.graph) - 1
	
	@property
	def total_stopping_sequence(self) -> Tuple[int]:
		# Tuple of the sequence from the initial value, to 1 (read-only attribute)
		return tuple(self.graph.nodes)
	
	@property
	def max(self) -> int:
		# Maximum value of the sequence (read-only attribute)
		return max(self.total_stopping_sequence)
	
	@property
	def max_rank(self) -> int:
		# Rank of the maximum value of the sequence (read-only attribute)
		return self.total_stopping_sequence.index(self.max)
	
	@property
	def graph_view(self) -> dict:
		# A dict-like structure representing the graph of the successive members of the current sequence (read-only attribute)
		return self.graph.adj
	
	@property
	def graph(self) -> nx.DiGraph:
		# The underlying, directional graph for the current Collatz sequence
		self._generate_graph()
		if self._populate_global_graph:
			if self._initial in self.single_graphs:
				return self.single_graphs[self._initial]
			else:
				self.single_graphs[self._initial] = self._graph = nx.path_graph(nx.shortest_path(self.global_graph, source=self._initial, target=1), create_using=nx.DiGraph)
				return self.single_graphs[self._initial]
		else:
			return self._graph
	
	def _generate_graph(self):
		"""Generate the internal, (directed) graph for the current sequence. Populate also the global graph if activated."""
		if not self._local_graph_completed:
			if self._populate_global_graph:
				for Un in self:
					if Un == self._initial:
						if Un in self.global_graph:
							self._local_graph_completed = True
							break
						else:
							self.global_graph.add_node(Un)
					else:
						if Un in self.global_graph:
							self.global_graph.add_edge(previous,Un)
							self._local_graph_completed = True
							break
						self.global_graph.add_edge(previous,Un)
					previous = Un
					
					if Un == 1:
						self._local_graph_completed = True
						break
			else:
				for Un in self:
					if Un == self._initial:
						self._graph.add_node(Un)
					else:
						self._graph.add_edge(previous,Un)
					previous = Un
					
					if Un == 1:
						self._local_graph_completed = True
						break
	
	
	@classmethod
	def _atomic_task_create_independent_global_graph0(cls, initials:Tuple[int]) -> nx.DiGraph:
		"""
		An elementary task executed by the workers.
		
		Here, the tasks consists of creating a global graph with the Collatz sequences beginning by the members of the `initials` tuple, WITHOUT populating the `global_graph` class property.
		
		Parameters:
			initials:
				Tuple of integers representing the initial values of the Collatz sequences building the global graph.
		
		Returns:
			networkx.DiGraph: A representation of the global graph of the Collatz sequences.
		
		Raises:
			ValueError:
				Raises if at least one member of `initials` is not a strictly positive integer.
		"""
		independent_global_graph = nx.DiGraph()
		for initial in initials:
			if not isinstance(initial, int) or initial<=0:
				raise ValueError(f"{initial} is not a strictly positive integer")
			collatz = cls(initial)
			for Un in collatz:
				if Un == collatz.initial_value:
						if Un in independent_global_graph:
							break
						else:
							independent_global_graph.add_node(Un)
				else:
					if Un in independent_global_graph:
						independent_global_graph.add_edge(previous,Un)
						break
					independent_global_graph.add_edge(previous,Un)
				previous = Un
				
				if Un == 1:
					break
		return independent_global_graph

	@classmethod
	def _atomic_task_create_independent_global_graph1(cls, initials:Tuple[int], global_edges:list[tuple], global_nodes:list[int]) -> None:
		"""
		An elementary task executed by the workers.
		
		Here, the tasks consists of creating a shared lists of edges and nodes, that will update the global graph later, with the Collatz sequences beginning by the members of the `initials` tuple, WITHOUT populating the `global_graph` class property.
		
		Parameters:
			initials:
				Tuple of integers representing the initial values of the Collatz sequences building the global graph.
			global_edges:
				Multiprocess shared list of 2-tuples representing the edges of the future global directional graph.
			global_nodes:
				Multiprocess shared list of the nodes of the future global directional graph.
		
		Raises:
			ValueError:
				Raises if at least one member of `initials` is not a strictly positive integer.
		"""
		local_edges = []
		local_nodes = []
		for initial in initials:
			if not isinstance(initial, int) or initial<=0:
				raise ValueError(f"{initial} is not a strictly positive integer")
			if initial not in global_nodes:
				collatz = cls(initial)
				for Un in collatz:
					if Un != initial:
						local_edges.append((previous,Un))
					if Un in local_nodes:
						break
					local_nodes.append(Un)
					previous = Un
					if Un == 1:
						break
		global_edges += local_edges
		global_nodes += local_nodes

	@classmethod
	def _atomic_task_create_independent_global_graph2(cls, initials:Tuple[int], queue:Queue) -> None:
		"""
		An elementary task executed by the workers.
		
		Here, the tasks consists of creating a directional graph with the Collatz sequences beginning by the members of the `initials` tuple, WITHOUT populating the `global_graph` class property, and sending the resulting graph to the parent process through the given multiprocessing.Queue.
		
		Parameters:
			initials:
				Tuple of integers representing the initial values of the Collatz sequences building the graph.
			queue (multiprocessing.Queue):
				Queue to send the list of edges to the parent process.
				
		Raises:
			ValueError:
				Raises if at least one member of `initials` is not a strictly positive integer.
		"""
		local_graph = nx.DiGraph()
		for initial in initials:
			if not isinstance(initial, int) or initial<=0:
				raise ValueError(f"{initial} is not a strictly positive integer")
			if initial not in local_graph:
				collatz = cls(initial)
				for Un in collatz:
					if Un == collatz.initial_value:
						if Un in local_graph:
							break
						else:
							local_graph.add_node(Un)
					else:
						if Un in local_graph:
							local_graph.add_edge(previous,Un)
							break
						local_graph.add_edge(previous,Un)
					previous = Un
					
					if Un == 1:
						break
		queue.put(local_graph)
	
	@classmethod
	def generate_global_graph(cls, max_initial_value:int, min_initial_value:int = 1, excludes:list[int] = [], reverse:bool = False, parallel:bool = False, parallel_algo:int = 0) -> nx.DiGraph:
		"""Generate a global graph gathering all Collatz sequences with initial values from `min_initial_value` to `max_initial_value`, without those beginning by members of `excludes`.
		
		To populate the graph, this function temporarly instanciates the needed Syracuse objects, that can cause lot of memory consumtion; nevertheless each instance are deleted before the initialization of the next one.
		The resulting graph is stored in the class attribute `global_graph`, and is returned by this function too for convenience.
		
		It is possible to switch to an alternative computation algorithm, using the ability of the computer/OS to execute simultaneous tasks. Depending on the hardware (ie: number of "cores" of the CPU), the benefit can be really interesting for a large range of values (the definition of "large" depends heavily on your configuration). For the most little ranges, it is better to use the classical, sequential approach.
		
		The available types of parallel computation are the following:
		
		| Type number | Description                                                             |
		| ----------- | ----------------------------------------------------------------------- |
		| 0           | Usage of `multiprocessing.Pool` objects.                                |
		| 1           | Usage of `multiprocessing.Process` objects: each process computes a part of edges and nodes of the global graph as shared objects, and the global graph is built in the parent process once all edges and nodes are ready.|
		| 2           | Usage of `multiprocessing.Process` objects: each process computes its own part of the global graph, and the full global graph is later gathered in the parent process. |
		
		Parameters:
			min_initial_value:
				The minimal initial value of the proceeded sequences.
			max_initial_value:
				The maximal initial value of the proceeded sequences.
			excludes:
				list of sequences excluded in the range. No effect for members lower than `min_initial_value` or greater than `max_initial_value`.
			reverse:
				If `True`, generate the global graph by computing the constitutive sequences from the maximal to the minimal initial value. Only relevant for non-parallel computation (ie: if `parallel` is `False`).
			parallel:
				If `True`, activates the parallel computation algorithm, using pool of multiprocessing workers.
			parallel_algo:
				Type of algorithm used for the parallel computation. Only relevant when `parallel` is `True`.
		
		Returns:
			networkx.DiGraph: A representation of the global graph of the Collatz sequences.
		
		Raises:
			ValueError:
				Raises if `min_initial_value` or `max_initial_value` are not strictly positive integers, or if `max_initial_value` < `min_initial_value`.
		"""
		if not isinstance(max_initial_value, int) or (max_initial_value <= 0):
			raise ValueError("max_initial_value must be a strictly positive integer")
		elif not isinstance(min_initial_value, int) or (min_initial_value <= 0):
			raise ValueError("min_initial_value must be a strictly positive integer")
		elif max_initial_value < min_initial_value:
			raise ValueError("max_initial_value must be greater than min_initial_value")
		else:
			if parallel:
				if parallel_algo == 1:
					with Manager() as manager:
						global_edges = manager.list()
						global_nodes = manager.list()
						nb_workers = os.cpu_count() - 1
						initial_list = [x for x in range(min_initial_value, max_initial_value+1) if x not in excludes]
						distribution = [tuple(chunk) for chunk in more_itertools.distribute(nb_workers,initial_list)] # Usage of more_itertools.distribute() because the order of elements in a chunk is unimportant
						workers = []
						for chunk in distribution:
							workers.append(Process(target=cls._atomic_task_create_independent_global_graph1, args=(chunk,global_edges,global_nodes)))
						for worker in workers:
							worker.start()
						for worker in workers:
							worker.join()
						cls.global_graph.update(edges=global_edges)

				elif parallel_algo == 2:
					with Manager() as manager:
						nb_workers = os.cpu_count()
						initial_list = [x for x in range(min_initial_value, max_initial_value+1) if x not in excludes]
						distribution = [tuple(chunk) for chunk in more_itertools.distribute(nb_workers,initial_list)] # Usage of more_itertools.distribute() because the order of elements in a chunk is unimportant
						workers = []
						queue = manager.Queue()
						for chunk in distribution:
							workers.append(Process(target=cls._atomic_task_create_independent_global_graph2, args=(chunk,queue)))
						for worker in workers:
							worker.start()
						for index, _ in enumerate(range(nb_workers)):
							graph = queue.get()
							cls.global_graph.update(graph)
				
				else:
					with Pool() as pool: # Number of worker processes: os.cpu_count() (default)
						initial_list = [x for x in range(min_initial_value, max_initial_value+1) if x not in excludes]
						distribution = [tuple(chunk) for chunk in more_itertools.distribute(os.cpu_count(),initial_list)] # Usage of more_itertools.distribute() because the order of elements in a chunk is unimportant
						result_iterable = pool.imap_unordered(cls._atomic_task_create_independent_global_graph0, distribution, 1)
						for graph in  result_iterable:
							cls.global_graph.update(graph)
			else:
				if reverse:
					for initial in range(max_initial_value, min_initial_value-1, -1):
						if initial not in excludes:
							cls(initial, populate_global_graph = True)
				else:
					for initial in range(min_initial_value, max_initial_value+1):
						if initial not in excludes:
							cls(initial, populate_global_graph = True)
			return cls.global_graph
	
	
	@classmethod
	def _atomic_task_total_stopping_time(cls, start:int) -> int:
		"""
		An elementary task executed by the workers.
		
		Here, the tasks consists of computing the total stopping time for the `Syracuse(start)` sequence.
		"""
		return cls(start).total_stopping_time
	
	@classmethod
	def total_stopping_times_range(cls, max_initial_value:int, min_initial_value:int = 1, parallel:bool = False) -> Tuple[int]:
		"""Generate the tuple of the total stopping times of all Collatz sequences with initial values from `min_initial_value` to `max_initial_value`.
		
		It is possible to switch to an alternative computation algorithm, using the ability of the computer/OS to execute simultaneous tasks. Depending on the hardware (ie: number of "cores" of the CPU), the benefit can be really interesting for a large range of values (the definition of "large" depends heavily on your configuration). For the most little ranges, it is better to use the classical, sequential approach.
			
		Parameters:
			min_initial_value:
				The minimal initial value of the proceeded sequences
			max_initial_value:
				The maximal initial value of the proceeded sequences
			parallel:
				If True, activates the parallel computation algorithm, using pool of multiprocessing workers
		
		Returns:
			The ordered total stopping times of the Collatz sequences
		
		Raises:
			ValueError:
				Raises if `min_initial_value` or `max_initial_value` are not strictly positive integers, or if `max_initial_value` < `min_initial_value`.
		"""

		if not isinstance(max_initial_value, int) or (max_initial_value <= 0):
			raise ValueError("max_initial_value must be a strictly positive integer")
		elif not isinstance(min_initial_value, int) or (min_initial_value <= 0):
			raise ValueError("min_initial_value must be a strictly positive integer")
		elif max_initial_value < min_initial_value:
			raise ValueError("max_initial_value must be greater than min_initial_value")
		else:
			if parallel:
				with Pool() as pool: # Number of worker processes: os.cpu_count() (default)
					# Leave Python computes the chunksize (see https://github.com/python/cpython/blob/3.11/Lib/multiprocessing/pool.py#L481 for details)
					return tuple(pool.map(cls._atomic_task_total_stopping_time, range(min_initial_value, max_initial_value+1)))
			else:
				total_stopping_times_list = []
				for initial in range(min_initial_value, max_initial_value+1):
					total_stopping_times_list.append(cls(initial).total_stopping_time)
				return tuple(total_stopping_times_list)
	
	@classmethod
	def _atomic_task_max(cls, start:int) -> int:
		"""
		An elementary task executed by the workers.
		
		Here, the tasks consists of computing the maximum value for the `Syracuse(start)` sequence.
		"""
		return cls(start).max
	
	@classmethod
	def max_reached_values_range(cls, max_initial_value:int, min_initial_value:int = 1, parallel:bool = False) -> Tuple[int]:
		"""Generate the tuple of the maximal values reached in all Collatz sequences with initial values from `min_initial_value` to `max_initial_value`.
		
		It is possible to switch to an alternative computation algorithm, using the ability of the computer/OS to execute simultaneous tasks. Depending on the hardware (ie: number of "cores" of the CPU), the benefit can be really interesting for a large range of values (the definition of "large" depends heavily on your configuration). For the most little ranges, it is better to use the classical, sequential approach.
			
		Parameters:
			min_initial_value:
				The minimal initial value of the proceeded sequences
			max_initial_value:
				The maximal initial value of the proceeded sequences
			parallel:
				If True, activates the parallel computation algorithm, using pool of multiprocessing workers
		
		Returns:
			A tuple with the ordered maximal reached values of the Collatz sequences
		
		Raises:
			ValueError:
				Raises if `min_initial_value` or `max_initial_value` are not strictly positive integers, or if `max_initial_value` < `min_initial_value`.
		"""
		if not isinstance(max_initial_value, int) or (max_initial_value <= 0):
			raise ValueError("max_initial_value must be a strictly positive integer")
		elif not isinstance(min_initial_value, int) or (min_initial_value <= 0):
			raise ValueError("min_initial_value must be a strictly positive integer")
		elif max_initial_value < min_initial_value:
			raise ValueError("max_initial_value must be greater than min_initial_value")
		else:
			if parallel:
				with Pool() as pool: # Number of worker processes: os.cpu_count() (default)
					# Leave Python computes the chunksize (see https://github.com/python/cpython/blob/3.11/Lib/multiprocessing/pool.py#L481 for details)
					return tuple(pool.map(cls._atomic_task_max, range(min_initial_value, max_initial_value+1)))
			else:
				max_reached_values_list = []
				for initial in range(min_initial_value, max_initial_value+1):
					max_reached_values_list.append(cls(initial).max)
				return tuple(max_reached_values_list)
	
	@classmethod
	def reset_global_graph(cls) -> None:
		"""
		Reset all graphs recorded in the class level: the global graph and the single_graphs list.
		"""
		cls.global_graph = nx.DiGraph()
		cls.single_graphs = {}
		

class CompressedSyracuse(Syracuse):
	"""
	A compressed Collatz sequence instanciated with a given initial value.
	
	This class provides the same parameters, attributes and methods as the inherited Syracuse class.
	"""
	
	def __iter__(self):
		Un = self._initial
		yield Un # Don't forget to yield the initial value !
		while True:
			Un = (Un*3+1)//2 if Un%2 else Un//2
			yield Un


"""This package provides some tools to deal with the Collatz conjecture.

The Collatz conjecture, also known as the "3n + 1 problem" or the "Syracuse problem", is an unsolved hypothesis in mathematics that concerns a sequence of operations applied to a positive integer. The conjecture states that no matter what the starting value of this integer is, the sequence will always eventually reach the value 1.

More specifically, the Collatz conjecture states that if one takes a positive integer n and applies the following function:

- If n is even, divide it by 2.
- If n is odd, multiply it by 3 and add 1.

Then, repeat the process with the resulting value, applying the function over and over again until the value of n eventually reaches 1.

!!! example
	
	If one takes n = 6, the sequence for the Collatz conjecture would be:
	6, 3, 10, 5, 16, 8, 4, 2, 1.

Although the conjecture has been computationally tested for extremely large values, it remains unproven to this day.

Since 3n+1 is even whenever n is odd, it is possible to define a "shortcut" (or "compressed") form for the Collatz function:

- If n is even, divide it by 2.
- If n is odd, multiply it by 3, add 1 **and divide the whole result by 2**.	
	
This definition yields smaller values for the stopping time and total stopping time without changing the overall dynamics of the process.

See the related [article on Wikipedia](https://en.wikipedia.org/wiki/Collatz_conjecture) for deeper information.

The package provides the following modules:

| Module    | Description                                                             |
| --------- | ----------------------------------------------------------------------- |
| `core`    | Provides classes to define Collatz sequences                            |
| `drawing` | Provides tool functions to render in various ways the Collatz sequences |
| `sequences` | Provides well-known sequences related to Collatz sequences |

It is important to notice that the `core` module can be used without being imported explicitly. However, all other modules **need** to be imported explicitly.

```pycon
>>> import syracuse # import all objects exposed by the core module
>>> import syracuse.drawing # import objects exposed by the drawing module
>>> import syracuse.sequences # import objects exposed by the sequences module
```

"""
from .core import Syracuse, CompressedSyracuse



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['syracuse']

package_data = \
{'': ['*']}

install_requires = \
['gmpy2>=2.1.5,<3.0.0',
 'matplotlib>=3.6,<3.7',
 'more-itertools>=9.1.0,<10.0.0',
 'networkx[default]>=3.0,<4.0',
 'numpy>=1.24,<1.25',
 'pydot>=1.4.2,<2.0.0']

setup_kwargs = {
    'name': 'syracuse',
    'version': '0.4.0',
    'description': 'The Syracuse problem (aka Collatz conjecture) library',
    'long_description': '![PyPI - Python Version](https://img.shields.io/pypi/pyversions/syracuse)\n![GitHub release (latest by date)](https://img.shields.io/github/v/release/devfred78/syracuse)\n![GitHub](https://img.shields.io/github/license/devfred78/syracuse)\n![GitHub issues](https://img.shields.io/github/issues/devfred78/syracuse)\n![GitHub pull requests](https://img.shields.io/github/issues-pr/devfred78/syracuse)\n\n![Graph of the first 26 compressed Collatz sequences](https://github.com/devfred78/syracuse/blob/main/assets/graph_compressed_syracuse_26.png?raw=True)\n*(Graph of the first 26 compressed Collatz sequences)*\n\n# Syracuse\n\nsyracuse, The Syracuse problem (aka Collatz conjecture) library.\n\n## About the project\n\nThe syracuse library provides a way to generate Collatz sequences (either "normal" or compressed forms) and some functions to deal with.\n\n### The Collatz conjecture\n\nThe Collatz conjecture, also known as the "3n + 1 problem" or the "Syracuse problem", is an unsolved hypothesis in mathematics that concerns a sequence of operations applied to a positive integer. The conjecture states that no matter what the starting value of this integer is, the sequence will always eventually reach the value 1.\n\nMore specifically, the Collatz conjecture states that if one takes a positive integer n and applies the following function:\n\n- If n is even, divide it by 2.\n- If n is odd, multiply it by 3 and add 1.\n\nThen, repeat the process with the resulting value, applying the function over and over again until the value of n eventually reaches 1.\n\n> For example, if one takes n = 6, the sequence for the Collatz conjecture would be:\n> 6, 3, 10, 5, 16, 8, 4, 2, 1.\n\nAlthough the conjecture has been computationally tested for extremely large values, it remains unproven to this day.\n\n### The compressed form\n\n*(definition partially inspired by the [relevant article on Wikipedia](https://en.wikipedia.org/wiki/Collatz_conjecture))*\n\nSince 3n+1 is even whenever n is odd, one may instead use the "compressed" form of the Collatz function:\n\n- if n is even, divide it by 2.\n- if n is odd, multiply it by 3, add 1 **and divide the result by 2**.\n\nThis definition yields smaller values for the stopping time and total stopping time without changing the overall dynamics of the process. \n\n> For example, consider the Collatz sequence for the number 17:\n> 17, 52, 26, 13, 40, 20, 10, 5, 16, 8, 4, 2, 1\n>\n> The compressed form of this sequence would be:\n> 17, 26, 13, 20, 10, 5, 8, 4, 2, 1\n\nThe compressed form of the Collatz sequence has been the subject of much research, as it provides insights into the behavior of the original sequence and can be used to search for cycles and other patterns. However, the Collatz conjecture itself remains unproven, and the compressed form of the sequence does not provide a proof or disproof of the conjecture.\n\n### Why the name "syracuse" ?\n\nThe Collatz conjecture is also known as the Syracuse problem because it was first introduced by Lothar Collatz in a 1950 paper, where he credited his colleague, mathematician Helmut Hasse, with calling it the Syracuse problem. The name "Syracuse" refers to the University of Syracuse, where Collatz was working at the time.\n\nThe name "Syracuse problem" has since become a common alternative to "Collatz conjecture", especially in French and Italian-speaking countries, where it is often referred to as the "problème de Syracuse" or "problema di Syracuse", respectively.\n\nRegardless of the name used, the conjecture remains one of the most famous and challenging open problems in mathematics, and it continues to attract the attention of mathematicians and computer scientists around the world.\n\n## Getting started\n\n### Prerequisites\n\nOf course, syracuse cannot run without Python ! More precisely, it requires at least the 3.11 version of our beloved language.\n\nsyracuse depends on the following packages. The installation of syracuse should install automatically those packages if they are missing on your system. If it fails, you can install them individually:\n\n* networkx: version 3.0.0 or above\n\n\t```sh\n\tpip install networkx["default"]\n\t```\n\n* matplotlib: version 3.6.0 or above\n\n\t```sh\n\tpip install matplotlib\n\t```\n\n* numpy: version 1.24.0 or above\n\n\t```sh\n\tpip install numpy\n\t```\n\n* pydot: version 1.4.2 or above\n\n\t```sh\n\tpip install pydot\n\t```\n\n* more-itertools: version 9.1.0 or above\n\n\t```sh\n\tpip install more-itertools\n\t```\n\n### Installation\n\nInstall from PyPi with:\n\n```sh\npip install syracuse\n```\n\nAs an alternative, you can download the `*.whl` file from the last [release on the syracuse Github repository](https://github.com/devfred78/syracuse/releases), and execute the following command (replace "X.Y.Z" by the right version number):\n\n```sh\npip install syracuse-X.Y.Z-py3-none-any.whl\n```\n\n## Usage\n\nBasic usage:\n\n```pycon\n>>> # Import the core library\n>>> import syracuse\n\n>>> # Create a Syracuse sequence with 27 as initial value\n>>> my_sequence = syracuse.Syracuse(27)\n\n>>> # Stopping time\n>>> my_sequence.stopping_time\n96\n>>> # Total stopping time\n>>> my_sequence.total_stopping_time\n111\n>>> # Maximum reached value\n>>> my_sequence.max\n9232\n>>> # Full sequence (until 1)\n>>> my_sequence.total_stopping_sequence\n(27, 82, 41, 124, 62, 31, 94, 47, 142, 71, 214, 107, 322, 161,\n 484, 242, 121, 364, 182, 91, 274, 137, 412, 206, 103, 310, 155,\n 466, 233, 700, 350, 175, 526, 263, 790, 395, 1186, 593, 1780,\n 890, 445, 1336, 668, 334, 167, 502, 251, 754, 377, 1132, 566,\n 283, 850, 425, 1276, 638, 319, 958, 479, 1438, 719, 2158, 1079,\n 3238, 1619, 4858, 2429, 7288, 3644, 1822, 911, 2734, 1367, 4102,\n 2051, 6154, 3077, 9232, 4616, 2308, 1154, 577, 1732, 866, 433,\n 1300, 650, 325, 976, 488, 244, 122, 61, 184, 92, 46, 23, 70, 35,\n 106, 53, 160, 80, 40, 20, 10, 5, 16, 8, 4, 2, 1)\n\n>>> # Import the sequences library\n>>> import syracuse.sequences\n\n>>> # Create an iterator over the records of the max values (A006884)\n>>> # with all possible optimizations\n>>> max_records = syracuse.sequences.max_value_records(15)\n\n>>> # Iterate over the first 30 records\n>>> for _ in range(30):\n...\t\tprint(f"{next(max_records)}, ", end="", flush=True)\n...\n1, 2, 3, 7, 15, 27, 255, 447, 639, 703, 1819, 4255, 4591, 9663,\n20895, 26623, 31911, 60975, 77671, 113383, 138367, 159487, 270271,\n665215, 704511, 1042431, 1212415, 1441407, 1875711, 1988859,\n\n>>> # Import the drawing library\n>>> import syracuse.drawing\n\n>>> # Display the full sequence in a Matplotlib window\n>>> syracuse.drawing.single_sequence_to_matplotlib_figure(my_sequence, test=True)\n```\n\n![Matplotlib graphical view of the successive values of the sequence Syracuse(27)](https://github.com/devfred78/syracuse/blob/main/assets/graph_sequence_27_reduced.png?raw=True)\n\nFor deeper explanations, please refer to the [documentation](https://devfred78.github.io/syracuse/).\n\n## Contributing\n\nContributions are what make the open source community such an amazing place to learn, inspire, and create. Any contributions you make are **greatly appreciated**.\n\nIf you have a suggestion that would make this better, please fork the repo and create a pull request. You can also simply open an issue with the tag "enhancement" or "bug", according to whether you want to share a proposal of a new function, or to record an anomaly.\n\nDon\'t forget to give the project a star! Thanks again!\n\n1. Fork the Project\n2. Create your Feature Branch (`git checkout -b feature/AmazingFeature`)\n3. Commit your Changes (`git commit -m \'Add some AmazingFeature\'`)\n4. Push to the Branch (`git push origin feature/AmazingFeature`)\n5. Open a Pull Request\n\n## License\n\nDistributed under the MIT License. Check out [LICENSE.md](https://github.com/devfred78/syracuse/blob/main/LICENSE.md) file for more information.\n\n## Acknowledgments\n\nI would like greatfully to thank:\n\nGraphviz [authors](https://graphviz.org/) for this impressive graph visualization software, especially for the creation of the useful [DOT Langage](https://graphviz.org/doc/info/lang.html).\n\n[The Matplotlib development team](https://matplotlib.org/) for providing a very powerful library "for creating static, animated, and interactive visualizations in Python".\n\nNumPy [community](https://numpy.org/) for this fundamental tool to be used as a priority if you want make serious scientific computations with Python.\n\n[MkDocs](https://www.mkdocs.org/), [mkdocstrings](https://mkdocstrings.github.io/) and [Material for MkDocs](https://squidfunk.github.io/mkdocs-material/) for giving the ability to create in a very simple way an accessible documentation for either tiny or big projects.\n\n[Make a README](https://www.makeareadme.com/), [Sayan Mondal](https://medium.com/swlh/how-to-make-the-perfect-readme-md-on-github-92ed5771c061), [Hillary Nyakundi](https://www.freecodecamp.org/news/how-to-write-a-good-readme-file/) and [othneildrew](https://github.com/othneildrew/Best-README-Template) for providing very interesting materials to write good README files (far better than I can write by myself !).\n\n[Choose an open source license](https://choosealicense.com/) for helping to choose the best suitable license for this project.\n\n[Semantic Versioning](https://semver.org/) for providing clear specifications for versioning projects.\n\n[Real Python](https://realpython.com/) for contributing really increasing skills in Python for everyone, novices or veterans.\n\n[GitHub](https://github.com/) for hosting this project, and helping to share it.\n\n[Pypi](https://pypi.org/) for providing a very convenient way to share modules and package to the entire Python community.\n\nAnd, of course, all the former, current and further contributors of this project !',
    'author': 'devfred78',
    'author_email': '97401939+devfred78@users.noreply.github.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/devfred78/syracuse',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.11,<3.12',
}


setup(**setup_kwargs)



"""
This module provides several well-known sequences related to Collatz sequences.

Unless explicitely mentioned, all sequences provided here implement the [iterator protocol](https://docs.python.org/3/library/stdtypes.html#iterator-types). That is, they can be used in `for` loops, and as parameter for the built-in `iter()` and `next()` functions. It also means that only one item is kept in memory at a time, the next items being generated on demand or lazily. Unlike container types like lists or dictionnaries, data are not stored but forgotten once the next item is yield. By the way, iterators are very memory-efficient and can process infinite data streams, like most of the sequences generated here.

The available sequences are the following:

| sequence | [OEIS](https://oeis.org/) reference | Description |
| -------- | ----------------------------------- | ----------- |
| total_stopping_time_records | A006877 | Sequence of starting values of Collatz sequences with a total stopping time longer than of any smaller starting value. |
| max_value_records | A006884 | Sequence of starting values of Collatz sequences with maximum values higher than of any smaller starting value. |


"""

from collections import namedtuple
from collections.abc import Iterator
from itertools import count

from gmpy2 import bit_scan1

Record = namedtuple("Record", "rank value")

class total_stopping_time_records(Iterator):
	"""
	Sequence of starting values of Collatz sequences with a total stopping time longer than of any smaller starting value. OEIS reference: [A006877](https://oeis.org/A006877).
	
	!!! example
		
		6 is included in this sequence because the total stopping time of the Collatz sequence beginning by 6 (that is, 8) is greater than total stopping times of all Collatz sequences beginning by 1, 2, 3, 4 and 5.
	
	In order to speedup the computation, it is possible to apply optional optimizations. Those optimizations are combinable (by addition of their numbers) for a greater efficiency. Deeper details (including mathematical demontrations) are available in [this article from T. Leavens and M. Vermeulen](https://oeis.org/A006877/a006877_1.pdf).
	
	The available sorts of optimizations are the following (the speedup factor is calculated with the first 50 records. It is only indicative, since it may differ on your own configuration):
	
	| Number | Name | Description | Speedup factor |
	| ------ | ---- | ----------- | -------------- |
	| 0      | None | No optimization | 1.0 |
	| 1      | Even numbers | Due to the fact that `Syracuse(2k).total_stopping_time` = `Syracuse(k).total_stopping_time + 1`, it is easy to predict the next even candidate for this sequence, and then it is not necessary to compute the Collatz sequences for the other even numbers below this candidate. | 1.9 |
	| 2      | k mod 6 = 5 | If the remainder of the division of the initial value by 6 is 5, then this cannot be a record for the total stopping time. | 1.2 |
	| 4      | *a posteriori* cutoff | It is possible to stop the iteration process of a Collatz sequence before its end, by comparing the current iterate value with all previous records and the number of steps necessary to reach it. **This optimization needs to store all items previously computed, making this iterator less memory-efficient**. | 2.0 |
	| 8      | make_odd | Speed up the iterations of the Collatz sequences by replacing the successive divisions by 2, by only one step. Be aware that using this "optimization" alone is absolutly not efficient (as you can see, the speedup factor is more or less 1). However, it is a real "booster" when associated with other optimization algorithms. For instance, the *a posteriori* cutoff is boosted with a speedup factor of 3.3 when used together with the "make_odd" one. | 1.0 |
	
	Parameters:
		optimization:
			Type of optimization(s) applied. No optimization by default.
	"""
	
	# Optimizations
	NO_OPT = 0
	EVEN_OPT = 1
	KMOD6_OPT = 2
	APOST_OPT = 4
	MKODD_OPT = 8
	
	def __init__(self, optimization:int = 0):
		self.optimization = optimization
		self.previous_record = -1
		if self.optimization & self.EVEN_OPT:
			self.previous_rank = 0
		if self.optimization & self.APOST_OPT:
			self.records = list()
	
	def __next__(self):
		n = 1
		while True:
			n_discarded = False
			if (self.optimization & self.KMOD6_OPT) and (n%6 == 5):
				# n is not a record in total stopping time
				value = self.previous_record
				n_discarded = True

			if (self.optimization & self.EVEN_OPT) and not n%2:
				if n == self.previous_rank * 2:
					value = self._steps_mkodd(n) if (self.optimization & self.MKODD_OPT) else self._steps_apost(n) if (self.optimization & self.APOST_OPT) else self._steps(n)
					n_discarded = True
				else:
					# n is not a record in total stopping time
					value = self.previous_record
					n_discarded = True
			
			if not n_discarded:
				value = self._steps_mkodd(n) if (self.optimization & self.MKODD_OPT) else self._steps_apost(n) if (self.optimization & self.APOST_OPT) else self._steps(n)
			
			if value is not None and (value > self.previous_record):
				self.previous_record = value
				if self.optimization & self.EVEN_OPT:
					self.previous_rank = n
				if self.optimization & self.APOST_OPT:
					self.records.append(Record(rank=n, value=value))
				return n
			n += 1
	
	def _steps(self, n:int) -> int:
		"""
		The total stopping time of the Collatz sequence beginning by `n`.
		
		Since the values of the Collatz sequence are not stored, the usage of a syracuse.Syracuse() instance is not necessary.
		
		Parameters:
			n:
				Initial value of the Collatz sequence for which the total stopping time is calculated.
		
		Returns:
			The total stopping time.
		
		Raises:
			ValueError:
				Raises if `n` is not a strictly positive integer.
		"""
		
		if n<=0 or not isinstance(n, int):
			raise ValueError(f"{n} is not a strictly positive integer")

		steps = 0
		
		while n>1:
			n = n*3+1 if n%2 else n//2
			steps += 1
		
		return steps
	
	def _steps_apost(self, n:int) -> int|None:
		"""
		The total stopping time of the Collatz sequence beginning by `n`, using the *a posteriori* cutoff optimization.
		
		If the *a posteriori* cutoff algorithm concludes that the current Collatz sequence is not a record, then it is immediatly dropped of, and the function returns `None`.
		
		Parameters:
			n:
				Initial value of the Collatz sequence for which the total stopping time is calculated.
		
		Returns:
			The total stopping time, or `None` if the iteration process has been stopped before its end.
		
		Raises:
			ValueError:
				Raises if `n` is not a strictly positive integer.
		"""
		
		if n<=0 or not isinstance(n, int):
			raise ValueError(f"{n} is not a strictly positive integer")

		steps = 0
		
		while n>1:
			if n%2:
				n = n*3+1
				steps += 1
			else:
				n = n//2
				steps += 1
				if len(self.records) > 0:
					upper_records = [record for record in self.records if n < record.rank]
					for record in upper_records:
						if steps + record.value <= self.previous_record:
							return
					
		return steps
	
	def _steps_mkodd(self, n:int) -> int|None:
		"""
		The total stopping time of the Collatz sequence beginning by `n`, using the "make_odd" optimization.
		
		This method implements also the optional, *a posteriori* cutoff optimization.
		
		Parameters:
			n:
				Initial value of the Collatz sequence for which the total stopping time is calculated.
		
		Returns:
			The total stopping time, or `None` if the iteration process has been stopped before its end (if the *a posteriori* cutoff optimization is activated).
		
		Raises:
			ValueError:
				Raises if `n` is not a strictly positive integer.
		"""
		
		if n<=0 or not isinstance(n, int):
			raise ValueError(f"{n} is not a strictly positive integer")

		steps = 0
		
		while n>1:
			if n%2:
				n = n*3+1
				steps += 1
			
			# At this step, n is necessarily even
			n>>=(p := bit_scan1(n)) # successive divisions by 2
			steps += p
			
			# Optional, *a posteriori* cutoff optimization
			if (self.optimization & self.APOST_OPT):
				if len(self.records) > 0:
					upper_records = [record for record in self.records if n < record.rank]
					for record in upper_records:
						if steps + record.value <= self.previous_record:
							return

		return steps

class max_value_records(Iterator):
	"""
	Sequence of starting values of Collatz sequences with maximum values higher than of any smaller starting value. OEIS reference: [A006884](https://oeis.org/A006884).
	
	!!! example
	
		7 is included in this sequence because the maximum value of the Collatz sequence beginning by 7 (that is, 52) is greater than maximum values of all Collatz sequences beginning by 1, 2, 3, 4, 5 and 6.
	
	In order to speedup the computation, it is possible to apply optional optimizations. Those optimizations are combinable (by addition of their numbers) for a greater efficiency. Deeper details (including mathematical demontrations) are available in [this article from T. Leavens and M. Vermeulen](https://oeis.org/A006877/a006877_1.pdf).
	
	The available sorts of optimizations are the following (the speedup factor is calculated with the first 40 records. It is only indicative, since it may differ on your own configuration):
	
	| Number | Name | Description | Speedup factor |
	| ------ | ---- | ----------- | -------------- |
	| 0      | None | No optimization | 1.0 |
	| 1      | Odd numbers only | Except 2, there is no even record for max value. So it is not necessary to compute Collatz sequences starting by even values. | 2.0 |
	| 2      | k mod 6 = 5 | If the remainder of the division of the initial value by 6 is 5, then this cannot be a record for the max value. | 1.2 |
	| 4      | *a posteriori* cutoff | A record, if found, is a max value appearing before the iteration process has fallen under the initial value. Hence, once the initial value is reached, it is possible to stop the iteration process. | 8.0 |
	| 8      | make_odd | Speed up the iterations of the Collatz sequences by replacing the successive divisions by 2, by only one step. | 1.9 |
	
	Parameters:
		optimization:
			Type of optimization(s) applied. No optimization by default.
	"""
	
	# Optimizations
	NO_OPT = 0
	ODD_OPT = 1
	KMOD6_OPT = 2
	APOST_OPT = 4
	MKODD_OPT = 8
	
	def __init__(self, optimization:int = 0):
		self.optimization = optimization
		self.previous_record = -1
	
	def __next__(self):
		n = 1
		while True:
			if n%2 == 0:
				if (self.optimization & self.ODD_OPT) and (n!=2):
					# n is not a record in max value
					value = self.previous_record
				else:
					value = self._max_mkodd(n) if (self.optimization & self.MKODD_OPT) else self._max(n)
			elif (self.optimization & self.KMOD6_OPT) and (n%6 == 5):
				# n is not a record in max value
				value = self.previous_record
			else:
				value = self._max_mkodd(n) if (self.optimization & self.MKODD_OPT) else self._max(n)
			
			if value > self.previous_record:
				self.previous_record = value
				return n
			
			n += 1
	
	def _max(self, n:int) -> int:
		"""
		The maximum value of the Collatz sequence beginning by `n`.
		
		Since the values of the Collatz sequence are not stored, the usage of a syracuse.Syracuse() instance is not necessary.
		If the *a posteriori* cutoff optimization is enabled, the search for the max value is stopped when the iteration falls below the initial value.
		
		Parameters:
			n:
				Initial value of the Collatz sequence for which the maximum value is calculated.
		
		Returns:
			The absolute maximum value or the the maximum value reached before fallen below the initial value (if the *a posteriori* cutoff optimization is enabled).
		
		Raises:
			ValueError:
				Raises if `n` is not a strictly positive integer.
		"""
		
		if n<=0 or not isinstance(n, int):
			raise ValueError(f"{n} is not a strictly positive integer")
		
		initial_value = n
		max_value = n
		
		while n>1:
			if n%2:
				n = n*3+1
				max_value = max(max_value, n)
				if (self.optimization & self.APOST_OPT) and (n < initial_value):
					break
			else:
				n = n//2
		
		return max_value
	
	def _max_mkodd(self, n:int) -> int:
		"""
		The maximum value of the Collatz sequence beginning by `n`, using the "make_odd" optimization.
		
		This method implements also the optional, *a posteriori* cutoff optimization.
		
		Parameters:
			n:
				Initial value of the Collatz sequence for which the maximum value is calculated.
		
		Returns:
			The absolute maximum value or the the maximum value reached before fallen below the initial value (if the *a posteriori* cutoff optimization is enabled).
		
		Raises:
			ValueError:
				Raises if `n` is not a strictly positive integer.
		"""
		if n<=0 or not isinstance(n, int):
			raise ValueError(f"{n} is not a strictly positive integer")
		
		initial_value = n
		max_value = n
		
		while n>1:
			if n%2:
				n = n*3+1
			# At this step, n is necessarily even
			max_value = max(max_value, n)
			n>>=bit_scan1(n) # successive divisions by 2
			if (self.optimization & self.APOST_OPT) and (n < initial_value):
				break
		
		return max_value